const float4x4 View;
const float4x4 Proj;

#ifndef XBOX360
const float4 Params0[72];
const float4 Params1[72];
const float4 Params2[72];
#endif

const float ShadowDepthOffset;
const float3 LightColors[24];

struct VSInput
{
    float4 Position : POSITION0;
    float4 Color : COLOR0;
};

struct VSOutputDepth
{
    float4 Position : POSITION0;
    float4 Color : COLOR0;
    float Depth : TEXCOORD0;
    float2 TexCoord : TEXCOORD1;
    float3 Normal : TEXCOORD2;
};

struct VSOutputEdge
{
    float4 Position : POSITION0;
    float4 Color : COLOR0;
};

struct PSInputDepth
{
    float4 Color : COLOR0;
    float Depth : TEXCOORD0;
    float2 TexCoord : TEXCOORD1;
    float3 Normal : TEXCOORD2;
};

struct PSInputEdge
{
    float4 Color : COLOR0;
};

float3 TransformByQuaternion(float3 position, float4 quaternion)
{
    float4 Q = quaternion;
    float3 v = position;
    return
    (2.0f * Q.w * Q.w - 1.0f) * v +
    (2.0f * dot( v, Q.xyz ) * Q.xyz) +
    (2.0f * Q.w * cross(Q.xyz, v));
}

static float3 lightRates[5] = {0.3, 0.3, 0.3, 0.7, 0.7, 0.7, 1, 1, 1, 0.7, 0.7, 0.7, 0.3, 0.3, 0.3};

VSOutputDepth TransformDepth(VSInput input, float4 param0, float4 param1, float4 param2:COLOR1)
{
    VSOutputDepth output;
    float3 offset = param0.xyz;
    float3 scale = param0.w;
    float3 vtpos = TransformByQuaternion(input.Position * scale, param1);
    float4 pos = float4(offset + vtpos, 1);
    float4 vpos = mul(pos, View);
    output.Position = mul(vpos, Proj);
    output.Color.rgba = input.Color.rgba * input.Color.a + param2.rgba * (1 - input.Color.a);
    output.Color.rgb *= (sin((vpos.z + ShadowDepthOffset) * 0.05) + 1) * 0.5 + 1;
    float3 lightColor = 0;
    int li = 4 - offset.z / 32 - 2;
    for (int i = 0; i < 5; i++, li++)
        lightColor += LightColors[li] * lightRates[i];
    output.Color.rgb *= 0.5 - smoothstep(0, 500, length(lightColor)) * 0.5;
    output.Color.rgb += lightColor * 0.002;
    output.Depth.x = -vpos.z;
    output.TexCoord.x = input.Position.x * 5;
    output.TexCoord.y = (input.Position.y + input.Position.z) * 2;
    output.Normal.xyz = normalize(pos.xyz * 2 + vtpos.xyz);
    return output;
}

VSOutputEdge TransformBloom(VSInput input, float4 param0, float4 param1, float4 param2:COLOR1)
{
    VSOutputEdge output;
    float3 offset = param0.xyz;
    float3 scale = param0.w;
    float3 vtpos = TransformByQuaternion(input.Position * scale, param1);
    float4 pos = float4(offset + vtpos, 1);
    float4 vpos = mul(pos, View);
    output.Position = mul(vpos, Proj);
    float depth = -vpos.z;
    float a = (1 - smoothstep(36, 256, depth)) * smoothstep(20, 36, depth);
    output.Color.rgb = input.Color.rgb * a + param2.rgb * (1 - a);
    output.Color.a = input.Color.a * a;
    return output;
}

VSOutputEdge TransformEdge(VSInput input, float4 param0, float4 param1, float4 param2:COLOR1)
{
    VSOutputEdge output;
    float3 offset = param0.xyz;
    float3 scale = param0.w;
    float3 vtpos = TransformByQuaternion(input.Position * scale, param1);
    float4 pos = float4(offset + vtpos, 1);
    float4 vpos = mul(pos, View);
    output.Position = mul(vpos, Proj);
    float depth = -vpos.z;
    float a = (1 - smoothstep(36, 256, depth)) * smoothstep(20, 36, depth);
    output.Color.rgb = input.Color.rgb * a + param2.rgb * (1 - a);
    output.Color.a = 1 - a * 0.2;
    return output;
}

VSOutputEdge TransformParticle(VSInput input, float4 param0, float4 param1, float4 param2:COLOR1)
{
    VSOutputEdge output;
    float3 offset = param0.xyz;
    float3 scale = param0.w;
    float3 vtpos = TransformByQuaternion(input.Position * scale, param1);
    float4 pos = float4(offset + vtpos, 1);
    float4 vpos = mul(pos, View);
    output.Position = mul(vpos, Proj);
    output.Color.r = param2.r * (1 + sin(vtpos.x * 2 + vtpos.y * 1 + vtpos.z * 3) * 0.5);
    output.Color.g = param2.g * (1 + sin(vtpos.x * 3 + vtpos.y * 2 + vtpos.z * 1) * 0.5);
    output.Color.b = param2.b * (1 + sin(vtpos.x * 1 + vtpos.y * 3 + vtpos.z * 2) * 0.5);
    output.Color.a = param2.a * smoothstep(20, 36, -vpos.z);
    return output;
}

VSOutputEdge TransformParticleBloom(VSInput input, float4 param0, float4 param1, float4 param2:COLOR1)
{
    VSOutputEdge output;
    float3 offset = param0.xyz;
    float3 scale = param0.w;
    float3 vtpos = TransformByQuaternion(input.Position * scale, param1);
    float4 pos = float4(offset + vtpos, 1);
    float4 vpos = mul(pos, View);
    output.Position = mul(vpos, Proj);
    output.Color.r = param2.r * (1 + sin(vtpos.x * 2 + vtpos.y * 1 + vtpos.z * 3) * 0.5);
    output.Color.g = param2.g * (1 + sin(vtpos.x * 3 + vtpos.y * 2 + vtpos.z * 1) * 0.5);
    output.Color.b = param2.b * (1 + sin(vtpos.x * 1 + vtpos.y * 3 + vtpos.z * 2) * 0.5);
    float depth = -vpos.z;
    float a = (1 - smoothstep(50, 640, depth)) * smoothstep(20, 50, depth);
    output.Color.a = param2.a * a;
    return output;
}

VSOutputEdge TransformLetter(VSInput input, float4 param0, float4 param1, float4 param2:COLOR1)
{
    VSOutputEdge output;
    float3 offset = param0.xyz;
    float3 scale = param0.w;
    float3 vtpos = TransformByQuaternion(input.Position * scale, param1);
    float4 pos = float4(offset + vtpos, 1);
    float4 vpos = mul(pos, View);
    output.Position = mul(vpos, Proj);
    output.Color = param2;
    return output;
}

#ifndef XBOX360
VSOutputDepth TransformDepthSI(VSInput input, float index:TEXCOORD0)
{
    return TransformDepth(input, Params0[index], Params1[index], Params2[index]);
}

VSOutputEdge TransformBloomSI(VSInput input, float index:TEXCOORD0)
{
    return TransformBloom(input, Params0[index], Params1[index], Params2[index]);
}

VSOutputEdge TransformEdgeSI(VSInput input, float index:TEXCOORD0)
{
    return TransformEdge(input, Params0[index], Params1[index], Params2[index]);
}

VSOutputEdge TransformParticleSI(VSInput input, float index:TEXCOORD0)
{
    return TransformParticle(input, Params0[index], Params1[index], Params2[index]);
}

VSOutputEdge TransformParticleBloomSI(VSInput input, float index:TEXCOORD0)
{
    return TransformParticleBloom(input, Params0[index], Params1[index], Params2[index]);
}

VSOutputEdge TransformLetterSI(VSInput input, float index:TEXCOORD0)
{
    return TransformLetter(input, Params0[index], Params1[index], Params2[index]);
}

#else

const int VertexCount;
static float4 param0;
static float4 param1;
static float4 param2;

VSInput CreateVSInput(int index)
{
    int vi = (index + 0.5) % VertexCount;
    int ii = (index + 0.5) / VertexCount;
    float4 pos;
    float4 color;
    asm
    {
        vfetch pos, vi, position0
        vfetch color, vi, color0
        vfetch param0, ii, texcoord0
        vfetch param1, ii, texcoord1
        vfetch param2, ii, color1
    };
    VSInput input;
    input.Position = pos;
    input.Color = color;
    return input;
}

VSOutputDepth TransformDepthHI(int index:INDEX)
{
    VSInput input = CreateVSInput(index);
    return TransformDepth(input, param0, param1, param2);
}

VSOutputEdge TransformBloomHI(int index:INDEX)
{
    VSInput input = CreateVSInput(index);
    return TransformBloom(input, param0, param1, param2);
}

VSOutputEdge TransformEdgeHI(int index:INDEX)
{
    VSInput input = CreateVSInput(index);
    return TransformEdge(input, param0, param1, param2);
}

VSOutputEdge TransformParticleHI(int index:INDEX)
{
    VSInput input = CreateVSInput(index);
    return TransformParticle(input, param0, param1, param2);
}

VSOutputEdge TransformParticleBloomHI(int index:INDEX)
{
    VSInput input = CreateVSInput(index);
    return TransformParticleBloom(input, param0, param1, param2);
}

VSOutputEdge TransformLetterHI(int index:INDEX)
{
    VSInput input = CreateVSInput(index);
    return TransformLetter(input, param0, param1, param2);
}
#endif

static float3 lightVector = {1, 0, 0};

float4 FillDepth(PSInputDepth input):COLOR0
{
    float4 color;
    color.rgb = input.Color.rgb * ((dot(input.Normal, lightVector) + 2) / 3);
    color.a = input.Color.a * (0.75 + sin(input.TexCoord.x + input.TexCoord.y) * 0.25) * (1 - smoothstep(32, 640, input.Depth.x));
    return color;
}

float4 Fill(float4 color:COLOR0):COLOR0
{
    return color;
}

sampler samplerState; 

float4 FillSampler(float2 texCoord:TEXCOORD0):COLOR0
{
    return tex2D(samplerState, texCoord);
}

static float2 offsets[12] =
{
    -0.3, -0.4, -0.8, -0.1, -0.7, 0.5, -0.2, 0.6, 1.0, -0.1, 0.5, -0.5,
    0.5, 0.8, 0.2, -0.9, 0.5, 0.1, 0.9, 0.4, -0.3, -1.0, -0.8, -0.6,
};

float4 FillSamplerBloom(float2 texCoord:TEXCOORD0):COLOR0
{
    float4 sum = tex2D(samplerState, texCoord);
    for (int i = 0; i < 12; i++)
        sum += tex2D(samplerState, texCoord + 0.01 * offsets[i]);
    sum /= 5;
    return sum;
}

technique DepthTech
{
    pass Pass1
    {
#ifdef XBOX360
        VertexShader = compile vs_3_0 TransformDepthHI();
#else
        VertexShader = compile vs_2_0 TransformDepthSI();
#endif
        PixelShader = compile ps_2_0 FillDepth();
    }
}

technique BloomTech
{
    pass Pass1
    {
#ifdef XBOX360
        VertexShader = compile vs_3_0 TransformBloomHI();
#else
        VertexShader = compile vs_2_0 TransformBloomSI();
#endif
        PixelShader = compile ps_1_1 Fill();
    }
}

technique EdgeTech
{
    pass Pass1
    {
#ifdef XBOX360
        VertexShader = compile vs_3_0 TransformEdgeHI();
#else
        VertexShader = compile vs_2_0 TransformEdgeSI();
#endif
        PixelShader = compile ps_1_1 Fill();
    }
}

technique ParticleTech
{
    pass Pass1
    {
#ifdef XBOX360
        VertexShader = compile vs_3_0 TransformParticleHI();
#else
        VertexShader = compile vs_2_0 TransformParticleSI();
#endif
        PixelShader = compile ps_1_1 Fill();
    }
}

technique ParticleBloomTech
{
    pass Pass1
    {
#ifdef XBOX360
        VertexShader = compile vs_3_0 TransformParticleBloomHI();
#else
        VertexShader = compile vs_2_0 TransformParticleBloomSI();
#endif
        PixelShader = compile ps_1_1 Fill();
    }
}

technique LetterTech
{
    pass Pass1
    {
#ifdef XBOX360
        VertexShader = compile vs_3_0 TransformLetterHI();
#else
        VertexShader = compile vs_2_0 TransformLetterSI();
#endif
        PixelShader = compile ps_1_1 Fill();
    }
}

technique FillTech
{
    pass Pass1
    {
        PixelShader = compile ps_1_1 FillSampler();
    }
}

technique FillBloomTech
{
    pass Pass1
    {
        PixelShader = compile ps_2_0 FillSamplerBloom();
    }
}
