/*
 * $Id: Enemy.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Abagames.Util;

namespace Abagames.GearToyGear
{
    /// <summary>
    /// Small size enemies.
    /// </summary>
    public class EnemyPool : ActorPool<Enemy>
    {
        private Random random;
        private Frame frame;
        private GameState gameState;
        private Field field;
        private BulletPool bullets;
        private Player player;
        private PillarPool pillars;
        private ParticlePool particles;
        private Sound sound;
        private PlayerHomingLaserPool playerHomingLasers;
        private GearEdgeShape edgeShape;

        public EnemyPool(int n, Frame frame, GameState gameState, Field field, BulletPool bullets, Player player, PillarPool pillars, ParticlePool particles, Sound sound)
            : base(n)
        {
            this.frame = frame;
            this.gameState = gameState;
            this.field = field;
            this.bullets = bullets;
            this.player = player;
            this.pillars = pillars;
            this.particles = particles;
            this.sound = sound;
            shape = new GearShape(frame, 1, 0.7f, 0.3f, new Color(240, 180, 180, 200), new Color(0, 0, 0, 200), true);
            edgeShape = new GearEdgeShape(frame, 1, 0.7f, 0.3f, 0.2f);
        }

        public void SetParams(PlayerHomingLaserPool playerHomingLasers)
        {
            this.playerHomingLasers = playerHomingLasers;
        }

        public void SetRandomSeed(int seed)
        {
            random = new Random(seed);
        }

        public void Add(EnemyType type, EnemyMotionType motionType)
        {
            Enemy a = new Enemy();
            a.Scale = type.Scale;
            a.Orientation = Quaternion.Identity;
            a.Color = type.Color;
            a.Ticks = 0;
            a.Motion.Type = motionType;
            a.Motion.Initialize();
            a.Pos = a.Motion.Pos;
            a.PPos = a.Pos;
            a.Type = type;
            a.FireInterval = a.FireTicks = type.FireInterval;
            a.Roll = 0;
            Add(ref a);
        }

        public override bool Update(ref Enemy a)
        {
            a.PPos = a.Pos;
            a.Motion.Update();
            a.Pos = a.Motion.Pos;
            Vector3 vel = (a.Pos - a.PPos);
            if (pillars.CheckHit(a.Pos))
            {
                particles.Add(40, a.Pos, vel * 2, 0.75f, a.Scale * 1.5f, 0.8f, 0.4f, 0.2f, 0.7f);
                return false;
            }
            a.Ticks++;
            if (a.Pos.Z < Field.FireBoundaryDepth)
                a.FireTicks -= Stage.GameSpeed;
            if (a.FireTicks <= 0)
            {
                Fire(ref a);
                a.FireTicks = a.Type.FireInterval;
            }
            a.Roll += 10.0f / a.FireTicks;
            a.Orientation = Quaternion.CreateFromYawPitchRoll(vel.X * 0.3f, vel.Y * 0.3f, a.Roll);
            return (a.Pos.Z < Field.FrontDepth && a.Pos.Z > Field.BackDepth);
        }

        private void Fire(ref Enemy a)
        {
            sound.PlaySe3D("Shot", a.Pos);
            if (a.Type.IsFiringStraight)
                bullets.Add(a.Pos, a.Type.FireSpeed);
            else
                bullets.Add(a.Pos, player.Pos, a.Type.FireSpeed);
        }

        public bool CheckHit(Vector3 pos)
        {
            bool isHit = false;
            for (int i = 0; i < actorCount;)
            {
                if (Vector3.Distance(Actors[i].Pos, pos) < Actors[i].Scale * 1.5f)
                {
                    Hit(i);
                    isHit = true;
                }
                else
                {
                    i++;
                }
            }
            return isHit;
        }

        public void Hit(int i)
        {
            particles.Add(20, Actors[i].Pos, 2, 20, Actors[i].Scale * 3, 1, 0.75f, 0.5f, 0.8f);
            Vector3 vel = (Actors[i].Pos - Actors[i].PPos);
            vel *= (2 + (float)random.NextDouble() * 2);
            particles.Add(80, Actors[i].Pos, vel, 0.5f, Actors[i].Scale * 2, 1, 0.75f, 0.5f, 0.6f);
            gameState.AddScore(100);
            sound.PlaySe3D("EnemyDestroyed", Actors[i].Pos);
            Remove(i);
        }

        public int GetNearest(float x, float y, float distance)
        {
            int ri = -1, i = 0;
            float  md = distance;
            ForEach(delegate(ref Enemy a)
            {
                float d = Math.Abs(a.Pos.X - x) + Math.Abs(a.Pos.Y - y);
                if (!a.IsLocked && d < md && a.Pos.Z < 0)
                {
                    ri = i;
                    md = d;
                }
                i++;
            });
            if (ri >= 0)
                Actors[ri].IsLocked = true;
            return ri;
        }

        public override void Draw(ref Enemy a)
        {
            Shape.AddInstance(a.Pos, a.Scale, a.Orientation, a.Color);
        }

        public void DrawEdge()
        {
            Draw(edgeShape);
        }

        public override void Remove(int i)
        {
            playerHomingLasers.OnEnemyRemoved(false, i);
            base.Remove(i);
        }
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct Enemy
    {
        public Vector3 Pos;
        public float Scale;
        public Quaternion Orientation;
        public Vector4 Color;
        public int Ticks;
        public EnemyMotion Motion;
        public Vector3 PPos;
        public float FireInterval;
        public float FireTicks;
        public float Roll;
        public EnemyType Type;
        public bool IsLocked;
    }

    public struct EnemyType
    {
        public bool IsFiringStraight;
        public float FireInterval;
        public float FireSpeed;
        public float Scale;
        public Vector4 Color;
    }
}
