/*
 * $Id: Frame.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.GamerServices;
using Abagames.Util;

namespace Abagames.GearToyGear
{
    /// <summary>
    /// A game state and effect parameters.
    /// </summary>
    public class Frame : Abagames.Util.Frame
    {
        private const string titleName = "GearToyGear";
        private const int versionNum = 10;
        private enum GameState
        {
            Title, InGame,
        };
        private Random random = new Random();
        private EffectParameter viewParam;
        private EffectParameter projParam;
        private Pad pad;
        private Title title;
        private ActorPools actors;
        private int pauseTicks;
        private GameState state;
        private bool backPressed;
        private bool startPressed;
        private bool fullScreenPressed;
        private Record record;
        private Replay replay;
        private Sound sound;
        private RenderTarget2D bloomSurface, edgeSurface;
        private int screenWidth, screenHeight;
        private SpriteBatch spriteBatch;

        public override void Initialize()
        {
            base.Initialize();
#if XBOX360
            game.Components.Add(new GamerServicesComponent(game));
#endif
        }

        public override void LoadContent()
        {
            window.Title = titleName;
            effectFileName = @"Shaders\Base";
            defaultTechniqueName = "EdgeTech";
            base.LoadContent();
            viewParam = Effect.Parameters["View"];
            projParam = Effect.Parameters["Proj"];
            spriteBatch = new SpriteBatch(Gd);
            screenWidth = window.ClientBounds.Width;
            screenHeight = window.ClientBounds.Height;
            ProjMatrix = Matrix.CreatePerspectiveFieldOfView((float)Math.PI / 2.0f, (float)screenWidth / screenHeight, 0.1f, 100000.0f);
            projParam.SetValue(ProjMatrix);
            bloomSurface = CreateRenderTarget(Gd, 1, Gd.PresentationParameters.BackBufferFormat);
            edgeSurface = CreateRenderTarget(Gd, 1, Gd.PresentationParameters.BackBufferFormat);
            Shape.Initialize(this);
            Letter.Initialize(content, this);
            pad = new Pad();
            replay = new Replay();
            record = new Record(titleName + "_record.dat", titleName, versionNum, 10000);
            record.Load();
            sound = new Sound();
            sound.Initialize();
            title = new Title(this, pad, record);
            actors = new ActorPools(this, content, pad, replay, record, sound);
            StartTitle();
        }

        private RenderTarget2D CreateRenderTarget(GraphicsDevice device, int numberLevels, SurfaceFormat surface)
        {
            if (!GraphicsAdapter.DefaultAdapter.CheckDeviceFormat(DeviceType.Hardware,
                GraphicsAdapter.DefaultAdapter.CurrentDisplayMode.Format, TextureUsage.None,
                QueryUsages.None, ResourceType.RenderTarget, surface))
                surface = device.DisplayMode.Format;
            return new RenderTarget2D(device,
                device.PresentationParameters.BackBufferWidth,
                device.PresentationParameters.BackBufferHeight,
                numberLevels, surface,
                device.PresentationParameters.MultiSampleType,
                device.PresentationParameters.MultiSampleQuality);

        }

        public override void SetRenderState(RenderState renderState) {
            renderState.CullMode = CullMode.None;
            renderState.DepthBufferEnable = true;
            renderState.AlphaBlendEnable = true;
            renderState.SourceBlend = Blend.SourceAlpha;
            renderState.DestinationBlend = Blend.InverseSourceAlpha;
        }

        public void StartTitle()
        {
            state = GameState.Title;
            Start(replay.RandomSeed);
            if (replay.IsAvailable)
                actors.StartReplay();
            else
                actors.StartTitle();
            title.Initialize();
            sound.StopBgm();
        }

        public void StartGame()
        {
            state = GameState.InGame;
            int randomSeed = random.Next();
            replay.RandomSeed = randomSeed;
            Start(randomSeed);
            actors.StartRecord();
            pauseTicks = -1;
            startPressed = true;
        }

        public void StartGameOver()
        {
            sound.FadeoutBgm();
        }

        public void Start(int randomSeed)
        {
            actors.Start(randomSeed);
            backPressed = true;
        }

        public override void Update()
        {
            record.Update();
            sound.Update();
            pad.Update();
            bool bp = false;
            if (pad.ButtonBack)
            {
                if (!backPressed)
                    bp = true;
                backPressed = true;
            }
            else
            {
                backPressed = false;
            }
            if (pad.ButtonFullScreen)
            {
                if (!fullScreenPressed)
                    graphics.ToggleFullScreen();
                fullScreenPressed = true;
            }
            else
            {
                fullScreenPressed = false;
            }
            switch (state)
            {
                case GameState.Title:
                    if (bp)
                    {
                        Exit();
                        return;
                    }
                    if (replay.IsAvailable)
                        UpdateInGame();
                    else
                        actors.UpdateField();
                    title.Update();
                    break;
                case GameState.InGame:
                    if (bp)
                    {
                        StartTitle();
                        return;
                    }
                    if (pad.ButtonStart)
                    {
                        if (!startPressed)
                        {
                            startPressed = true;
                            if (pauseTicks < 0)
                                pauseTicks = 0;
                            else
                                pauseTicks = -1;
                        }
                    }
                    else
                    {
                        startPressed = false;
                    }
                    if (pauseTicks >= 0)
                        pauseTicks++;
                    UpdateInGame();
                    break;
            }
        }

        private void UpdateInGame()
        {
            if (pauseTicks < 0)
                actors.Update();
        }

        public override void Draw()
        {
            if (state == GameState.Title && !replay.IsAvailable)
                DrawTitle();
            else
                DrawInGame();
        }

        private void DrawTitle()
        {
            ChangeTechnique("DepthTech");
            ClearScreen();
            actors.DrawField();
            ChangeTechnique("LetterTech");
            Gd.RenderState.DepthBufferEnable = false;
            title.Draw();
            actors.DrawScore();
            Letter.Draw();
            Gd.RenderState.DepthBufferEnable = true;
        }

        private void DrawInGame()
        {
            Gd.SetRenderTarget(0, edgeSurface);
            Gd.Clear(Color.TransparentBlack);
            actors.DrawEdge();
            Gd.SetRenderTarget(0, bloomSurface);
            Gd.Clear(Color.TransparentBlack);
            ChangeTechnique("ParticleBloomTech");
            actors.DrawParticle();
            ChangeTechnique("BloomTech");
            actors.DrawBloom();
            ChangeTechnique("DepthTech");
            Gd.SetRenderTarget(0, null);
            ClearScreen();
            actors.Draw();
            ChangeTechnique("ParticleTech");
            actors.DrawParticle();
            ChangeTechnique("LetterTech");
            Gd.RenderState.DepthBufferEnable = false;
            if (state == GameState.Title)
            {
                title.Draw();
                actors.DrawScore();
            }
            else
            {
                actors.DrawGameState();
            }
            if (pauseTicks >= 0 && pauseTicks % 60 < 30)
                Letter.Add("PAUSE", new Vector3(-3, -4, 0), 0.7f, Quaternion.Identity, 1);
            Letter.Draw();
            Gd.RenderState.DepthBufferEnable = true;
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.None);
            ChangeTechnique("FillTech");
            spriteBatch.Draw(edgeSurface.GetTexture(), Vector2.Zero, Color.White);
            spriteBatch.End();
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.None);
            ChangeTechnique("FillBloomTech");
            spriteBatch.Draw(bloomSurface.GetTexture(), Vector2.Zero, Color.White);
            spriteBatch.End();
            SetRenderState(Gd.RenderState);
        }

        private void ClearScreen()
        {
            Gd.Clear(new Color(Stage.BackgroundR, Stage.BackgroundG, Stage.BackgroundB));
        }

        public void LookAt(Vector3 from, Vector3 to, Vector3 up)
        {
            ViewMatrix = Matrix.CreateLookAt(from, to, up);
            viewParam.SetValue(ViewMatrix);
        }

        public int PauseTicks
        {
            get
            {
                return pauseTicks;
            }
        }

        public bool IsInGame
        {
            get
            {
                return (state == GameState.InGame);
            }
        }
    }
}
