/*
 * $Id: GameState.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using Microsoft.Xna.Framework;
using Abagames.Util;

namespace Abagames.GearToyGear
{
    /// <summary>
    /// Game state including score, number of players.
    /// </summary>
    public class GameState
    {
        private const int firstExtendScore = 1000000;
        private const int maxLeft = 9;
        private const int gameOverTicksBase = 300;
        private Frame frame;
        private Pad pad;
        private Record record;
        private Sound sound;
        private PlatePool plates;
        private int score;
        private int multiplier;
        private int extendScore, nextExtendScoreOffset;
        private int left;
        private bool isInGameOver;
        private int gameOverTicks;
        private bool isButtonPressed;
        private int ticks;
        private bool isAccelPressed;
        private int bgmIndex, bgmIndexChange;
        private Random random;

        public GameState(Frame frame, Pad pad, Record record, Sound sound, PlatePool plates)
        {
            this.frame = frame;
            this.pad = pad;
            this.record = record;
            this.sound = sound;
            this.plates = plates;
        }

        public void SetRandomSeed(int seed)
        {
            random = new Random(seed);
        }

        public void Initialize()
        {
            score = 0;
            multiplier = 1;
            left = 2;
            extendScore = nextExtendScoreOffset = firstExtendScore;
            ticks = 0;
            isInGameOver = false;
            isAccelPressed = false;
            bgmIndex = random.Next(bgmNames.Length);
            bgmIndexChange = random.Next(2) * 2 - 1;
        }

        private string[] bgmNames = { "Gtg1", "Gtg2", "Gtg3" };

        public void StartBgm()
        {
            if (!frame.IsInGame)
                return;
            sound.PlayBgm(bgmNames[bgmIndex]);
            bgmIndex += bgmIndexChange;
            if (bgmIndex >= bgmNames.Length)
                bgmIndex = 0;
            else if (bgmIndex < 0)
                bgmIndex = bgmNames.Length - 1;
        }

        public void FadeoutBgm()
        {
            sound.FadeoutBgm();
        }

        public void AddScore(int s)
        {
            if (isInGameOver)
                return;
            score += s * multiplier;
            if (score >= extendScore)
            {
                if (left < maxLeft)
                {
                    sound.PlaySe("Extend");
                    left++;
                }
                nextExtendScoreOffset += firstExtendScore;
                extendScore += nextExtendScoreOffset;
            }
        }

        public void AddScore(int s, Vector3 p, float scale)
        {
            if (isInGameOver)
                return;
            plates.Add((float)Math.Sqrt(p.X * p.X + p.Y * p.Y), (float)Math.Atan2(p.X, p.Y), p.Z, s * multiplier, scale);
            AddScore(s);
        }

        public void Miss()
        {
            left--;
            if (left < 0)
            {
                frame.StartGameOver();
                isInGameOver = true;
                gameOverTicks = gameOverTicksBase;
                isButtonPressed = true;
                if (frame.IsInGame)
                    record.RecordScore(score);
            }
        }

        public void Update()
        {
            ticks++;
            if (!isAccelPressed && ticks > 480)
                isAccelPressed = true;
            if (!isInGameOver)
                return;
            if (pad.ButtonAny)
            {
                if (!isButtonPressed)
                {
                    if (frame.IsInGame)
                    {
                        if (gameOverTicks < gameOverTicksBase - 30)
                            frame.StartGame();
                    }
                    else
                    {
                        frame.StartGame();
                    }
                    isButtonPressed = true;
                }
            }
            else
            {
                isButtonPressed = false;
            }
            gameOverTicks--;
            if (gameOverTicks <= 0)
                frame.StartTitle();
        }

        public void OnAccelPressed()
        {
            isAccelPressed = true;
        }

        public void Draw()
        {
            frame.LookAt(new Vector3(0, 0, 25), Vector3.Zero, new Vector3(0, 1, 0));
            Vector3 p = Letter.Add((int)multiplier, new Vector3(28, 20, 0), 0.7f, Quaternion.Identity, 1);
            p.Y += 0.2f;
            Letter.Add("X", p, 0.5f, Quaternion.Identity, 1);
            Letter.Add(score, new Vector3(28, -20, 0), 1, Quaternion.Identity, 1);
            Letter.Add(extendScore, new Vector3(28.5f, -17.5f, 0), 0.4f, Quaternion.Identity, 1);
            if (isInGameOver)
                Letter.Add("GAME OVER", new Vector3(-12, 0, 0), 1.5f - gameOverTicks * 0.001f, Quaternion.Identity, 1);
            else
                Letter.Add(left, new Vector3(-28, -20, 0), 1, Quaternion.Identity, 1);
            if (!isAccelPressed && ticks > 300)
                Letter.Add("PRESS RT TO SPEED UP. LT TO BREAK.", new Vector3(-15, 8, 0), 0.5f, Quaternion.Identity, 1);
        }

        public void DrawLastScore()
        {
            frame.LookAt(new Vector3(0, 0, 25), Vector3.Zero, new Vector3(0, 1, 0));
            Letter.Add(record.LastScore, new Vector3(28, -20, 0), 1, Quaternion.Identity, 1);
        }

        public int Multiplier
        {
            set
            {
                multiplier = value;
            }
        }

        public bool IsInGameOver
        {
            get
            {
                return isInGameOver;
            }
        }
    }
}
