/*
 * $Id: Particle.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Abagames.Util;

namespace Abagames.GearToyGear
{
    /// <summary>
    /// Particles.
    /// </summary>
    public class ParticlePool : ActorPool<Particle>
    {
        private const float MinimumScale = 1;
        private Vector3[] lightColors = new Vector3[24];
        private Frame frame;
        private Random random;

        public ParticlePool(int n, Frame frame)
            : base(n)
        {
            shape = new CubeShape(frame, 0.2f, 0.2f, 1);
            this.frame = frame;
        }

        public void SetRandomSeed(int seed)
        {
            random = new Random(seed);
        }

        public override void Clear()
        {
            for (int i = 0; i < lightColors.Length; i++)
            {
                lightColors[i].X = 0;
                lightColors[i].Y = 0;
                lightColors[i].Z = 0;
            }
            frame.Effect.Parameters["LightColors"].SetValue(lightColors);
            frame.Effect.CommitChanges();
            base.Clear();
        }

        public void Add(int n, Vector3 p, Vector3 vel, float velFluctuation, float scale, float r, float g, float b, float alpha)
        {
            Particle a = new Particle();
            for (int i = 0; i < n; i++)
            {
                a.Vel.X = vel.X * ((float)random.NextDouble() * velFluctuation + (1 - velFluctuation / 2));
                a.Vel.Y = vel.Y * ((float)random.NextDouble() * velFluctuation + (1 - velFluctuation / 2));
                a.Vel.Z = vel.Z * ((float)random.NextDouble() * velFluctuation + (1 - velFluctuation / 2));
                a.Pos.X = p.X + a.Vel.X;
                a.Pos.Y = p.Y + a.Vel.Y;
                a.Pos.Z = p.Z + a.Vel.Z;
                a.Orientation = Quaternion.Identity;
                a.TargetScale = scale;
                a.BaseScale = MinimumScale;
                a.Color.X = r * ((float)random.NextDouble() * 0.5f + 0.75f);
                a.Color.Y = g * ((float)random.NextDouble() * 0.5f + 0.75f);
                a.Color.Z = b * ((float)random.NextDouble() * 0.5f + 0.75f);
                a.Color.W = alpha;
                Add(ref a);
            }
        }

        public void Add(int n, Vector3 p, float velFluctuation, float velFluctuationZ, float scale, float r, float g, float b, float alpha)
        {
            Particle a = new Particle();
            for (int i = 0; i < n; i++)
            {
                a.Vel.X = ((float)random.NextDouble() - 0.5f) * velFluctuation;
                a.Vel.Y = ((float)random.NextDouble() - 0.5f) * velFluctuation;
                a.Vel.Z = ((float)random.NextDouble() - 0.5f) * velFluctuationZ;
                a.Pos.X = p.X + a.Vel.X;
                a.Pos.Y = p.Y + a.Vel.Y;
                a.Pos.Z = p.Z + a.Vel.Z;
                a.Orientation = Quaternion.Identity;
                a.TargetScale = scale;
                a.BaseScale = MinimumScale;
                a.Color.X = r * ((float)random.NextDouble() * 0.5f + 0.75f);
                a.Color.Y = g * ((float)random.NextDouble() * 0.5f + 0.75f);
                a.Color.Z = b * ((float)random.NextDouble() * 0.5f + 0.75f);
                a.Color.W = alpha;
                Add(ref a);
            }
        }

        public override void Update()
        {
            for (int i = 0; i < lightColors.Length; i++)
            {
                lightColors[i].X = 0;
                lightColors[i].Y = 0;
                lightColors[i].Z = 0;
            }
            base.Update();
            frame.Effect.Parameters["LightColors"].SetValue(lightColors);
            frame.Effect.CommitChanges();
        }

        public override bool Update(ref Particle a)
        {
            a.PPos = a.Pos;
            a.Pos += a.Vel;
            if (a.Pos.X * a.Pos.X + a.Pos.Y * a.Pos.Y > Tube.Radius * Tube.Radius)
            {
                Vector3 nv = new Vector3(-a.Pos.X, -a.Pos.Y, 0);
                if (Vector3.Dot(nv, a.Vel) < 0)
                {
                    nv.Normalize();
                    a.Vel = Vector3.Reflect(a.Vel, nv);
                }
            }
            a.BaseScale += (a.TargetScale - a.BaseScale) * 0.1f;
            Vector3 v = a.Pos - a.PPos;
            float vl = v.Length();
            a.Orientation.X = -v.Y;
            a.Orientation.Y = v.X;
            a.Orientation.Z = 0;
            a.Orientation.W = v.Z + vl;
            a.Orientation.Normalize();
            if (vl > 10)
                vl = 10;
            a.Scale = vl * a.BaseScale * 0.1f;
            a.Vel.Z -= 0.1f;
            a.Vel *= 0.98f;
            a.TargetScale *= 0.98f;
            a.Color.W *= 0.99f;
            int li = 4 - (int)(a.Pos.Z / 32);
            if (li >= 0 && li < lightColors.Length)
            {
                float lr = a.Scale * a.Color.W;
                lightColors[li].X += a.Color.X * lr;
                lightColors[li].Y += a.Color.Y * lr;
                lightColors[li].Z += a.Color.Z * lr;
            }
            return (a.Scale >= MinimumScale);
        }

        public override void Draw(ref Particle a)
        {
            Shape.AddInstance(a.Pos, a.Scale, a.Orientation, a.Color);
        }
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct Particle
    {
        public Vector3 Pos;
        public float Scale;
        public Quaternion Orientation;
        public Vector4 Color;
        public Vector3 Vel;
        public float TargetScale;
        public float BaseScale;
        public Vector3 PPos;
    }
}
