/*
 * $Id: Pillar.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Abagames.Util;

namespace Abagames.GearToyGear
{
    /// <summary>
    /// Obstacle pillars.
    /// </summary>
    public class PillarPool : ActorPool<Pillar>
    {
        private const float BASE_THICKNESS = 1.0f;
        private Frame frame;
        private Field field;
        private Player player;
        private Stage stage;
        private GearEdgeShape edgeShape;

        public PillarPool(int n, Frame frame, Field field, Player player)
            : base(n)
        {
            this.frame = frame;
            this.field = field;
            this.player = player;
            shape = new GearShape(frame, 1, 0.7f, BASE_THICKNESS, new Color(240, 240, 240, 255), new Color(100, 100, 100, 255), true);
            edgeShape = new GearEdgeShape(frame, 1, 0.7f, BASE_THICKNESS, 0.3f);
        }

        public void SetParams(Stage stage)
        {
            this.stage = stage;
        }

        public void Add(float radius, float angle, float size, float angleRate, float rollRatio)
        {
            Pillar a = CreateBasePillar();
            a.Pos.X = (float)Math.Sin(angle) * radius;
            a.Pos.Y = (float)Math.Cos(angle) * radius;
            a.Scale = size;
            a.Roll = Quaternion.CreateFromYawPitchRoll(0, 0, angleRate * rollRatio);
            a.Radius = radius;
            a.Angle = angle;
            a.AngleRate = angleRate;
            a.AppearanceTicks = -1;
            Add(ref a);
        }

        public void Add(float x, float y, float size, int appearanceTicks)
        {
            Pillar a = CreateBasePillar();
            a.Pos.X = x;
            a.Pos.Y = y;
            a.Scale = size;
            a.AppearanceTicks = appearanceTicks;
            if (appearanceTicks >= 0)
            {
                a.TargetScale = size;
                a.Scale = 0;
            }
            Add(ref a);
        }

        private Pillar CreateBasePillar()
        {
            Pillar a = new Pillar();
            a.Pos.Z = Field.BackDepth * 3;
            a.Orientation = Quaternion.Identity;
            a.Color.X = 0.5f;
            a.Color.Y = 1.0f;
            a.Color.Z = 1.0f;
            a.Color.W = 1.0f;
            return a;
        }

        public override bool Update(ref Pillar a)
        {
            if (a.AngleRate != 0)
            {
                a.Angle += a.AngleRate * Stage.GameSpeedSqrt;
                a.Pos.X = (float)Math.Sin(a.Angle) * a.Radius;
                a.Pos.Y = (float)Math.Cos(a.Angle) * a.Radius;
                a.Orientation *= a.Roll;
            }
            a.Pos.Z += Stage.PlayerDepthSpeed;
            if (a.Pos.Z > -BASE_THICKNESS * a.Scale && a.Pos.Z < BASE_THICKNESS * a.Scale + Stage.PlayerDepthSpeed)
            {
                Vector3 pp = player.Pos;
                Vector3 po = -player.Vel;
                po.Z = Stage.PlayerDepthSpeed;
                po /= 5;
                for (int i = 0; i < 4; i++)
                {
                    float ox = a.Pos.X - pp.X;
                    float oy = a.Pos.Y - pp.Y;
                    float oz = a.Pos.Z - pp.Z;
                    float s = a.Scale;
                    if (oz > -BASE_THICKNESS * s && oz < BASE_THICKNESS * s && ox * ox + oy * oy < s * s)
                        player.Destroy();
                    pp += po;
                }
            }
            if (a.AppearanceTicks >= 0)
            {
                if (a.AppearanceTicks < 30)
                {
                    a.Scale += (a.TargetScale - a.Scale) * 0.1f;
                    if (a.AppearanceTicks == 0)
                        a.Scale = a.TargetScale;
                }
                a.AppearanceTicks--;
            }
            return (a.Pos.Z < Field.FrontDepth);
        }

        public bool CheckHit(Vector3 pos)
        {
            for (int i = 0; i < actorCount; i++)
            {
                float oz = Actors[i].Pos.Z - pos.Z;
                if (oz > -BASE_THICKNESS * Actors[i].Scale && oz < BASE_THICKNESS * Actors[i].Scale)
                {
                    float ox = Actors[i].Pos.X - pos.X;
                    float oy = Actors[i].Pos.Y - pos.Y;
                    float s = Actors[i].Scale;
                    if (ox * ox + oy * oy < s * s)
                        return true;
                }
            }
            return false;
        }

        public override void Draw(ref Pillar a)
        {
            Shape.AddInstance(a.Pos, a.Scale, a.Orientation, a.Color);
        }

        public void DrawEdge()
        {
            Draw(edgeShape);
        }
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct Pillar
    {
        public Vector3 Pos;
        public float Scale;
        public Quaternion Orientation;
        public Vector4 Color;
        public Quaternion Roll;
        public float Radius;
        public float Angle;
        public float AngleRate;
        public float TargetScale;
        public int AppearanceTicks;
    }
}
