/*
 * $Id: Turret.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using System.Runtime.InteropServices;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Audio;
using Abagames.Util;

namespace Abagames.GearToyGear
{
    /// <summary>
    /// Enemy's turret.
    /// </summary>
    public class TurretPool : ActorPool<Turret>
    {
        private const float LaserBaseLength = 7.0f;
        private Frame frame;
        private Field field;
        private LaserPool lasers;
        private PillarPool pillars;
        private Player player;
        private ParticlePool particles;
        private Sound sound;
        private MiddleEnemyPool enemies;

        public TurretPool(int n, Frame frame, Field field, LaserPool lasers, PillarPool pillars, Player player, ParticlePool particles, Sound sound)
            :base(n)
        {
            this.frame = frame;
            this.field = field;
            this.lasers = lasers;
            this.pillars = pillars;
            this.player = player;
            this.particles = particles;
            this.sound = sound;
        }

        public void SetParams(MiddleEnemyPool enemies)
        {
            this.enemies = enemies;
        }

        public override void Clear()
        {
            ForEach(delegate(ref Turret a)
            {
                if (a.Cue != null)
                    a.Cue.Stop(AudioStopOptions.Immediate);
            });
            base.Clear();
        }

        public int Add(Vector3 pos, Vector3 vel, bool hasLaser)
        {
            Turret a = new Turret();
            a.Pos = pos;
            a.TargetVel = vel;
            a.Vel = vel * 14;
            if (hasLaser)
            {
                a.HasLaser = true;
                a.LaserLength = 0;
                a.LaserSpeed = 5.0f;
            }
            return Add(ref a);
        }

        public override bool Update(ref Turret a)
        {
            float pl = a.Pos.Length();
            bool isInTube = (pl < Tube.Radius * 1.2f);
            if (isInTube)
            {
                a.IsInTube = true;
                a.Vel += (a.TargetVel - a.Vel) * 0.2f;
            }
            else
            {
                if (a.IsInTube)
                    return false;
            }
            if (a.SourcePos.Z > Field.FireBoundaryDepth * 2)
                return false;
            a.Pos += a.Vel * Stage.GameSpeedSqrt;
            if (!a.HasLaser)
                return true;
            if (a.Cue == null)
            {
                a.Cue = sound.GetCue("Laser");
                a.Listener = new AudioListener();
                a.Emitter = new AudioEmitter();
                a.Cue.Apply3D(a.Listener, a.Emitter);
                a.Listener.Position = sound.ListenerPos;
                a.Cue.Play();
            }
            a.LaserLength += a.LaserSpeed * Stage.GameSpeed;
            Vector3 v = a.Pos - a.SourcePos;
            v.Normalize();
            Vector3 p = a.SourcePos;
            float l = a.LaserLength % LaserBaseLength;
            p += v * l;
            v *= LaserBaseLength;
            int i = (int)a.LaserLength;
            for (; ; l += LaserBaseLength)
            {
                if (p.X * p.X + p.Y * p.Y > Tube.Radius * Tube.Radius || pillars.CheckHit(p))
                {
                    a.LaserLength = l;
                    particles.Add(4, p, v, 0.2f, 15, 0.2f, 1, 0.4f, 0.8f);
                    a.Emitter.Position = p;
                    break;
                }
                if (p.Z > -10 && Vector3.Distance(p, player.Pos) < 5.0f)
                {
                    player.Destroy();
                }
                lasers.Add(p, v, LaserBaseLength * 0.8f);
                if (i % 12 == 0)
                    particles.Add(1, p, v, 0.1f, 7, 0.2f, 1, 0.4f, 0.8f);
                p += v;
                i++;
                if (l > a.LaserLength || p.Z > Field.FrontDepth)
                {
                    a.Emitter.Position = p;
                    break;
                }
            }
            a.Listener.Position = sound.ListenerPos;
            a.Cue.Apply3D(a.Listener, a.Emitter);
            return true;
        }

        public override void Remove(int i)
        {
            if (Actors[i].HasLaser)
            {
                Actors[i].Cue.Stop(AudioStopOptions.Immediate);
                Actors[i].Cue = null;
            }
            enemies.OnTurretRemoved(i);
            base.Remove(i);
        }
    }

    public struct Turret
    {
        public Vector3 Pos;
        public Vector3 TargetVel;
        public Vector3 Vel;
        public bool IsInTube;
        public bool HasLaser;
        public float LaserLength;
        public float LaserSpeed;
        public Vector3 SourcePos;
        public Cue Cue;
        public AudioListener Listener;
        public AudioEmitter Emitter;
    }
}
