/*
 * $Id: Shape.cs,v 1.1.1.1 2009/08/10 14:55:11 kenta Exp $
 *
 * Copyright 2009 Kenta Cho. Some rights reserved.
 */
using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Runtime.InteropServices;

namespace Abagames.Util
{
    /// <summary>
    /// 3D primitive shape base classes.
    /// </summary>
    public abstract class Shape
    {
        public const int MaxInstanceCount = 72;
        public const int MaxInstanceBytes = 4800;
        protected static Vector4[] params0 = new Vector4[MaxInstanceCount];
        protected static Vector4[] params1 = new Vector4[MaxInstanceCount];
        protected static Vector4[] params2 = new Vector4[MaxInstanceCount];
        protected static int instanceCount;
#if XBOX360
        public static WritableVertexBuffer writableVB;
        public static int writableVBOffset, writableVBInstanceSize;
#endif
        protected Frame frame;
        protected VertexDeclaration vdec;

        public static void Initialize(Frame frame)
        {
#if XBOX360
            writableVB = new WritableVertexBuffer(frame.Gd, MaxInstanceCount * MaxInstanceBytes, 3);
#endif
        }

#if XBOX360
        public static void SetToWritableVertexBuffer<T>(T[] instances, int index, int count) where T : struct
        {
            writableVB.Flip();
            writableVBOffset = writableVB.SetData<T>(instances, index, count);
            writableVBInstanceSize = Marshal.SizeOf(typeof(T));
            instanceCount = count;
        }
#endif

        public static void BeginAddInstance()
        {
            instanceCount = 0;
        }

        public static void AddInstance(Vector3 pos, float scale, Quaternion quaternion, Vector4 color)
        {
            params0[instanceCount].X = pos.X;
            params0[instanceCount].Y = pos.Y;
            params0[instanceCount].Z = pos.Z;
            params0[instanceCount].W = scale;
            params1[instanceCount].X = quaternion.X;
            params1[instanceCount].Y = quaternion.Y;
            params1[instanceCount].Z = quaternion.Z;
            params1[instanceCount].W = quaternion.W;
            params2[instanceCount].X = color.X;
            params2[instanceCount].Y = color.Y;
            params2[instanceCount].Z = color.Z;
            params2[instanceCount].W = color.W;
            instanceCount++;
        }

        public Shape(Frame frame)
        {
            this.frame = frame;
        }

        public abstract void Draw();
    }

    public abstract class PrimitiveShape : Shape
    {
        protected VertexPositionColor[] verts;
        protected int vertCount;

        public PrimitiveShape(Frame frame, int vc)
            : base(frame)
        {
            verts = new VertexPositionColor[vc];
            vertCount = vc;
            for (int i = 0; i < vertCount; i++)
                verts[i] = new VertexPositionColor(Vector3.Zero, Color.White);
#if XBOX360
            VertexElement[] extraElements = 
            {
                new VertexElement(1, 0, VertexElementFormat.Vector4,
                                        VertexElementMethod.Default,
                                        VertexElementUsage.TextureCoordinate, 0),
                new VertexElement(1, 16,VertexElementFormat.Vector4,
                                        VertexElementMethod.Default,
                                        VertexElementUsage.TextureCoordinate, 1),
                new VertexElement(1, 32,VertexElementFormat.Vector4,
                                        VertexElementMethod.Default,
                                        VertexElementUsage.Color, 1)
            };
#else
            short stream = 0;
            VertexElement[] extraElements =
            {
                new VertexElement(stream, (short)(VertexPositionColor.SizeInBytes),
                                  VertexElementFormat.Single, VertexElementMethod.Default,
                                  VertexElementUsage.TextureCoordinate, 0)
            };
#endif
            vdec = ExtendVertexDeclaration(frame.Gd, VertexPositionColor.VertexElements, extraElements);
        }

        private VertexDeclaration ExtendVertexDeclaration(GraphicsDevice graphicsDevice, VertexElement[] originalElements, VertexElement[] extraElements)
        {
            int length = originalElements.Length + extraElements.Length;
            VertexElement[] elements = new VertexElement[length];
            originalElements.CopyTo(elements, 0);
            extraElements.CopyTo(elements, originalElements.Length);
            return new VertexDeclaration(graphicsDevice, elements);
        }
    }

    public abstract class PrimitiveListShape : PrimitiveShape
    {
        protected int index = 0;
        protected ushort[] indices;
        protected VertexBuffer vertexBuffer;
        protected IndexBuffer indexBuffer;
        protected int primitiveCount;

        public PrimitiveListShape(Frame frame, int vc)
            : base(frame, vc)
        {
        }

        public PrimitiveListShape(Frame frame, int vc, int pc)
            : base(frame, vc)
        {
            primitiveCount = pc;
        }

        protected void CreateIndexBuffer()
        {
            indexBuffer = new IndexBuffer
                (frame.Gd, indices.Length * sizeof(ushort), BufferUsage.None, IndexElementSize.SixteenBits);
            indexBuffer.SetData<ushort>(indices);
        }

        protected void SetIndexedPrimitives()
        {
#if XBOX360
            frame.Effect.Parameters["VertexCount"].SetValue(vertCount);
#else
            frame.Effect.Parameters["Params0"].SetValue(params0);
            frame.Effect.Parameters["Params1"].SetValue(params1);
            frame.Effect.Parameters["Params2"].SetValue(params2);
#endif
            frame.Effect.CommitChanges();
#if XBOX360
            frame.Gd.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionColor.SizeInBytes);
            frame.Gd.Vertices[1].SetSource(writableVB.VertexBuffer, writableVBOffset, writableVBInstanceSize);
#else
            frame.Gd.Vertices[0].SetSource(vertexBuffer, 0, VertexPositionColor.SizeInBytes + sizeof(float));
#endif
            frame.Gd.Indices = indexBuffer;
            frame.Gd.VertexDeclaration = vdec;
        }

        public void Add(float x, float y, float z, byte r, byte g, byte b, byte a)
        {
            verts[index].Position.X = x;
            verts[index].Position.Y = y;
            verts[index].Position.Z = z;
            verts[index].Color = new Color(r, g, b, a);
            index++;
        }

        public void EndAdd()
        {
#if XBOX360
            vertexBuffer = new VertexBuffer(frame.Gd, vertCount * VertexPositionColor.SizeInBytes, BufferUsage.None);
            vertexBuffer.SetData<VertexPositionColor>(verts, 0, vertCount);
#else
            VertexBuffer ovb = new VertexBuffer(frame.Gd, vertCount * VertexPositionColor.SizeInBytes, BufferUsage.None);
            ovb.SetData<VertexPositionColor>(verts, 0, vertCount);
            byte[] ovd = new byte[vertCount * VertexPositionColor.SizeInBytes];
            ovb.GetData(ovd);
            byte[] vd = new byte[vertCount * (VertexPositionColor.SizeInBytes + sizeof(float)) * MaxInstanceCount];
            int op = 0;
            for (int i = 0; i < MaxInstanceCount; i++)
            {
                int sp = 0;
                byte[] iib = BitConverter.GetBytes((float)i);
                for (int j = 0; j < vertCount; j++)
                {
                    Array.Copy(ovd, sp, vd, op, VertexPositionColor.SizeInBytes);
                    sp += VertexPositionColor.SizeInBytes;
                    op += VertexPositionColor.SizeInBytes;
                    iib.CopyTo(vd, op);
                    op += sizeof(float);
                }
            }
            vertexBuffer = new VertexBuffer(frame.Gd, vd.Length, BufferUsage.None);
            vertexBuffer.SetData(vd);
#endif
        }

        public override void Draw()
        {
            if (instanceCount <= 0)
                return;
            SetIndexedPrimitives();
#if XBOX360
            frame.Gd.DrawIndexedPrimitives
                (PrimitiveType.TriangleList, 0, 0, vertCount, 0, primitiveCount * instanceCount);
#else
            frame.Gd.DrawIndexedPrimitives
                (PrimitiveType.TriangleList, 0, 0, vertCount * instanceCount, 0, primitiveCount * instanceCount);
            GraphicsDevice gd = vertexBuffer.GraphicsDevice;
            gd.Vertices[0].SetSource(null, 0, 0);
#endif
        }   
    }

    public class TriangleListShape : PrimitiveListShape
    {
        public TriangleListShape(Frame frame, int tc)
            : base(frame, tc * 3, tc)
        {
            indices = new ushort[vertCount * MaxInstanceCount];
            for (int i = 0; i < vertCount * MaxInstanceCount; i++)
                indices[i] = (ushort)i;
            CreateIndexBuffer();
        }
    }

    public class QuadListShape : PrimitiveListShape
    {
        private static ushort[] quadIndices = {0, 1, 3, 1, 2, 3};

        public QuadListShape(Frame frame, int qc)
            : base(frame, qc * 4, qc * 2)
        {
            indices = new ushort[qc * 6 * MaxInstanceCount];
            for (int i = 0; i < qc * MaxInstanceCount; i++)
                for (int j = 0; j < 6; j++)
                    indices[i * 6 + j] = (ushort)(i * 4 + quadIndices[j]);
            CreateIndexBuffer();
        }
    }

    public class CubeShape : PrimitiveListShape
    {
        private static ushort[] quadIndices = { 0, 1, 3, 1, 2, 3 };
        private static ushort[] cubeIndices = { 0, 1, 2, 3, 1, 0, 4, 5, 2, 1, 5, 6, 3, 2, 6, 7, 0, 3, 7, 4, 7, 6, 5, 4 };
        private static float[,] cubeOffsets = { { -1, -1, -1 }, { 1, -1, -1 }, { 1, 1, -1 }, { -1, 1, -1 }, { -1, -1, 1 }, { 1, -1, 1 }, { 1, 1, 1 }, { -1, 1, 1 } };

        public CubeShape(Frame frame, float x, float y, float z)
            : base(frame, 8, 12)
        {
            for (int i = 0; i < 8; i++)
            {
                Add(cubeOffsets[i, 0] * x, cubeOffsets[i, 1] * y, cubeOffsets[i, 2] * z, 0, 0, 0, 0);
            }
            EndAdd();
            indices = new ushort[36 * MaxInstanceCount];
            for (int i = 0; i < MaxInstanceCount; i++)
            {
                for (int j = 0; j < 6; j++)
                {
                    for (int k = 0; k < 6; k++)
                        indices[i * 36 + j * 6 + k] = (ushort)(i * 8 + cubeIndices[quadIndices[k] + j * 4]);
                }
            }
            CreateIndexBuffer();
        }
    }
}
